@@ -1,117 +1,89 @@
-# -*- coding=utf-8 -*-
+from __future__ import annotations
 
 import logging
-import operator
 import os
 import platform
 import sys
 from collections import defaultdict
+from pathlib import Path, WindowsPath
+from typing import (
+    Any,
+    Callable,
+    DefaultDict,
+    Dict,
+    Iterator,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
-import attr
 from packaging.version import Version
+from pydantic.v1 import Field, validator
 
-from ..compat import Path, lru_cache
-from ..environment import ASDF_DATA_DIR, MYPY_RUNNING, PYENV_ROOT, SYSTEM_ARCH
+from ..environment import ASDF_DATA_DIR, PYENV_ROOT, SYSTEM_ARCH
 from ..exceptions import InvalidPythonVersion
 from ..utils import (
-    RE_MATCHER,
-    _filter_none,
     ensure_path,
     expand_paths,
     get_python_version,
     guess_company,
     is_in_path,
     looks_like_python,
-    optional_instance_of,
     parse_asdf_version_order,
     parse_pyenv_version_order,
     parse_python_version,
-    path_is_pythoncore,
-    unnest,
 )
-from .mixins import BaseFinder, BasePath
-
-if MYPY_RUNNING:
-    from typing import (
-        Any,
-        Callable,
-        DefaultDict,
-        Dict,
-        Generator,
-        Iterator,
-        List,
-        Optional,
-        Tuple,
-        Type,
-        TypeVar,
-        Union,
-        overload,
-    )
-
-    from .._vendor.pep514tools.environment import Environment
-    from .path import PathEntry
-else:
-
-    def overload(f):
-        return f
-
+from .common import FinderBaseModel
+from .mixins import PathEntry
 
 logger = logging.getLogger(__name__)
 
 
-@attr.s(slots=True)
-class PythonFinder(BasePath, BaseFinder):
-    root = attr.ib(default=None, validator=optional_instance_of(Path), type=Path)
+class PythonFinder(PathEntry):
+    root: Path
     # should come before versions, because its value is used in versions's default initializer.
     #: Whether to ignore any paths which raise exceptions and are not actually python
-    ignore_unsupported = attr.ib(default=True, type=bool)
+    ignore_unsupported: bool = True
     #: Glob path for python versions off of the root directory
-    version_glob_path = attr.ib(default="versions/*", type=str)
+    version_glob_path: str = "versions/*"
     #: The function to use to sort version order when returning an ordered version set
-    sort_function = attr.ib(default=None)  # type: Callable
+    sort_function: Optional[Callable] = None
     #: The root locations used for discovery
-    roots = attr.ib(default=attr.Factory(defaultdict), type=defaultdict)
+    roots: Dict = Field(default_factory=lambda: defaultdict())
     #: List of paths discovered during search
-    paths = attr.ib(type=list)
-    #: shim directory
-    shim_dir = attr.ib(default="shims", type=str)
+    paths: List = Field(default_factory=lambda: list())
     #: Versions discovered in the specified paths
-    _versions = attr.ib(default=attr.Factory(defaultdict), type=defaultdict)
-    _pythons = attr.ib(default=attr.Factory(defaultdict), type=defaultdict)
+    _versions: Dict = Field(default_factory=lambda: defaultdict())
+    pythons_ref: Dict = Field(default_factory=lambda: defaultdict())
 
-    def __del__(self):
-        # type: () -> None
-        self._versions = defaultdict()
-        self._pythons = defaultdict()
-        self.roots = defaultdict()
-        self.paths = []
+    class Config:
+        validate_assignment = True
+        arbitrary_types_allowed = True
+        allow_mutation = True
+        include_private_attributes = True
+        # keep_untouched = (cached_property,)
 
     @property
-    def expanded_paths(self):
-        # type: () -> Generator
-        return (
-            path for path in unnest(p for p in self.versions.values()) if path is not None
-        )
+    def version_paths(self) -> Any:
+        return self._versions.values()
 
     @property
-    def is_pyenv(self):
-        # type: () -> bool
+    def is_pyenv(self) -> bool:
         return is_in_path(str(self.root), PYENV_ROOT)
 
     @property
-    def is_asdf(self):
-        # type: () -> bool
+    def is_asdf(self) -> bool:
         return is_in_path(str(self.root), ASDF_DATA_DIR)
 
-    def get_version_order(self):
-        # type: () -> List[Path]
+    def get_version_order(self) -> list[Path]:
         version_paths = [
             p
             for p in self.root.glob(self.version_glob_path)
             if not (p.parent.name == "envs" or p.name == "envs")
         ]
         versions = {v.name: v for v in version_paths}
-        version_order = []  # type: List[Path]
+        version_order = []
         if self.is_pyenv:
             version_order = [
                 versions[v] for v in parse_pyenv_version_order() if v in versions
@@ -129,8 +101,7 @@ class PythonFinder(BasePath, BaseFinder):
             version_order = version_paths
         return version_order
 
-    def get_bin_dir(self, base):
-        # type: (Union[Path, str]) -> Path
+    def get_bin_dir(self, base) -> Path:
         if isinstance(base, str):
             base = Path(base)
         if os.name == "nt":
@@ -138,16 +109,11 @@ class PythonFinder(BasePath, BaseFinder):
         return base / "bin"
 
     @classmethod
-    def version_from_bin_dir(cls, entry):
-        # type: (PathEntry) -> Optional[PathEntry]
-        py_version = None
+    def version_from_bin_dir(cls, entry) -> PathEntry | None:
         py_version = next(iter(entry.find_all_python_versions()), None)
         return py_version
 
-    def _iter_version_bases(self):
-        # type: () -> Iterator[Tuple[Path, PathEntry]]
-        from .path import PathEntry
-
+    def _iter_version_bases(self) -> Iterator[tuple[Path, PathEntry]]:
         for p in self.get_version_order():
             bin_dir = self.get_bin_dir(p)
             if bin_dir.exists() and bin_dir.is_dir():
@@ -157,8 +123,7 @@ class PythonFinder(BasePath, BaseFinder):
                 self.roots[p] = entry
                 yield (p, entry)
 
-    def _iter_versions(self):
-        # type: () -> Iterator[Tuple[Path, PathEntry, Tuple]]
+    def _iter_versions(self) -> Iterator[tuple[Path, PathEntry, tuple]]:
         for base_path, entry in self._iter_version_bases():
             version = None
             version_entry = None
@@ -193,15 +158,13 @@ class PythonFinder(BasePath, BaseFinder):
                 yield (base_path, entry, version_tuple)
 
     @property
-    def versions(self):
-        # type: () -> DefaultDict[Tuple, PathEntry]
+    def versions(self) -> DefaultDict[tuple, PathEntry]:
         if not self._versions:
             for _, entry, version_tuple in self._iter_versions():
                 self._versions[version_tuple] = entry
         return self._versions
 
-    def _iter_pythons(self):
-        # type: () -> Iterator
+    def _iter_pythons(self) -> Iterator:
         for path, entry, version_tuple in self._iter_versions():
             if path.as_posix() in self._pythons:
                 yield self._pythons[path.as_posix()]
@@ -211,59 +174,57 @@ class PythonFinder(BasePath, BaseFinder):
             else:
                 yield self.versions[version_tuple]
 
-    @paths.default
-    def get_paths(self):
-        # type: () -> List[PathEntry]
-        _paths = [base for _, base in self._iter_version_bases()]
+    @validator("paths", pre=True, always=True)
+    def get_paths(cls, v) -> list[PathEntry]:
+        if v is not None:
+            return v
+
+        _paths = [base for _, base in cls._iter_version_bases()]
         return _paths
 
     @property
-    def pythons(self):
-        # type: () -> DefaultDict[str, PathEntry]
-        if not self._pythons:
+    def pythons(self) -> dict:
+        if not self.pythons_ref:
             from .path import PathEntry
 
-            self._pythons = defaultdict(PathEntry)  # type: DefaultDict[str, PathEntry]
+            self.pythons_ref = defaultdict(PathEntry)
             for python in self._iter_pythons():
-                python_path = python.path.as_posix()  # type: ignore
-                self._pythons[python_path] = python
-        return self._pythons
+                python_path = python.path.as_posix()
+                self.pythons_ref[python_path] = python
+        return self.pythons_ref
 
     @pythons.setter
-    def pythons(self, value):
-        # type: (DefaultDict[str, PathEntry]) -> None
-        self._pythons = value
+    def pythons(self, value) -> None:
+        self.pythons_ref = value
 
-    def get_pythons(self):
-        # type: () -> DefaultDict[str, PathEntry]
+    def get_pythons(self) -> DefaultDict[str, PathEntry]:
         return self.pythons
 
-    @overload
     @classmethod
-    def create(cls, root, sort_function, version_glob_path=None, ignore_unsupported=True):
-        # type: (str, Callable, Optional[str], bool) -> PythonFinder
+    def create(
+        cls, root, sort_function, version_glob_path=None, ignore_unsupported=True
+    ) -> PythonFinder:
         root = ensure_path(root)
         if not version_glob_path:
             version_glob_path = "versions/*"
         return cls(
             root=root,
             path=root,
-            ignore_unsupported=ignore_unsupported,  # type: ignore
+            ignore_unsupported=ignore_unsupported,
             sort_function=sort_function,
             version_glob_path=version_glob_path,
         )
 
     def find_all_python_versions(
         self,
-        major=None,  # type: Optional[Union[str, int]]
-        minor=None,  # type: Optional[int]
-        patch=None,  # type: Optional[int]
-        pre=None,  # type: Optional[bool]
-        dev=None,  # type: Optional[bool]
-        arch=None,  # type: Optional[str]
-        name=None,  # type: Optional[str]
-    ):
-        # type: (...) -> List[PathEntry]
+        major: str | int | None = None,
+        minor: int | None = None,
+        patch: int | None = None,
+        pre: bool | None = None,
+        dev: bool | None = None,
+        arch: str | None = None,
+        name: str | None = None,
+    ) -> list[PathEntry]:
         """Search for a specific python version on the path. Return all copies
 
         :param major: Major python version to search for.
@@ -275,13 +236,13 @@ class PythonFinder(BasePath, BaseFinder):
         :param str arch: Architecture to include, e.g. '64bit', defaults to None
         :param str name: The name of a python version, e.g. ``anaconda3-5.3.0``
         :return: A list of :class:`~pythonfinder.models.PathEntry` instances matching the version requested.
-        :rtype: List[:class:`~pythonfinder.models.PathEntry`]
         """
 
         call_method = "find_all_python_versions" if self.is_dir else "find_python_version"
-        sub_finder = operator.methodcaller(
-            call_method, major, minor, patch, pre, dev, arch, name
-        )
+
+        def sub_finder(path):
+            return getattr(path, call_method)(major, minor, patch, pre, dev, arch, name)
+
         if not any([major, minor, patch, name]):
             pythons = [
                 next(iter(py for py in base.find_all_python_versions()), None)
@@ -289,8 +250,12 @@ class PythonFinder(BasePath, BaseFinder):
             ]
         else:
             pythons = [sub_finder(path) for path in self.paths]
+
         pythons = expand_paths(pythons, True)
-        version_sort = operator.attrgetter("as_python.version_sort")
+
+        def version_sort(py):
+            return py.as_python.version_sort
+
         paths = [
             p for p in sorted(pythons, key=version_sort, reverse=True) if p is not None
         ]
@@ -298,15 +263,14 @@ class PythonFinder(BasePath, BaseFinder):
 
     def find_python_version(
         self,
-        major=None,  # type: Optional[Union[str, int]]
-        minor=None,  # type: Optional[int]
-        patch=None,  # type: Optional[int]
-        pre=None,  # type: Optional[bool]
-        dev=None,  # type: Optional[bool]
-        arch=None,  # type: Optional[str]
-        name=None,  # type: Optional[str]
-    ):
-        # type: (...) -> Optional[PathEntry]
+        major: str | int | None = None,
+        minor: int | None = None,
+        patch: int | None = None,
+        pre: bool | None = None,
+        dev: bool | None = None,
+        arch: str | None = None,
+        name: str | None = None,
+    ) -> PathEntry | None:
         """Search or self for the specified Python version and return the first match.
 
         :param major: Major version number.
@@ -320,10 +284,12 @@ class PythonFinder(BasePath, BaseFinder):
         :returns: A :class:`~pythonfinder.models.PathEntry` instance matching the version requested.
         """
 
-        sub_finder = operator.methodcaller(
-            "find_python_version", major, minor, patch, pre, dev, arch, name
-        )
-        version_sort = operator.attrgetter("as_python.version_sort")
+        def sub_finder(obj):
+            return obj.find_python_version(major, minor, patch, pre, dev, arch, name)
+
+        def version_sort(path_entry):
+            return path_entry.as_python.version_sort
+
         unnested = [sub_finder(self.roots[path]) for path in self.roots]
         unnested = [
             p
@@ -333,8 +299,7 @@ class PythonFinder(BasePath, BaseFinder):
         paths = sorted(list(unnested), key=version_sort, reverse=True)
         return next(iter(p for p in paths if p is not None), None)
 
-    def which(self, name):
-        # type: (str) -> Optional[PathEntry]
+    def which(self, name) -> PathEntry | None:
         """Search in this path for an executable.
 
         :param executable: The name of an executable to search for.
@@ -347,26 +312,32 @@ class PythonFinder(BasePath, BaseFinder):
         return non_empty_match
 
 
-@attr.s(slots=True)
-class PythonVersion(object):
-    major = attr.ib(default=0, type=int)
-    minor = attr.ib(default=None)  # type: Optional[int]
-    patch = attr.ib(default=None)  # type: Optional[int]
-    is_prerelease = attr.ib(default=False, type=bool)
-    is_postrelease = attr.ib(default=False, type=bool)
-    is_devrelease = attr.ib(default=False, type=bool)
-    is_debug = attr.ib(default=False, type=bool)
-    version = attr.ib(default=None)  # type: Version
-    architecture = attr.ib(default=None)  # type: Optional[str]
-    comes_from = attr.ib(default=None)  # type: Optional[PathEntry]
-    executable = attr.ib(default=None)  # type: Optional[str]
-    company = attr.ib(default=None)  # type: Optional[str]
-    name = attr.ib(default=None, type=str)
+class PythonVersion(FinderBaseModel):
+    major: int = 0
+    minor: Optional[int] = None
+    patch: Optional[int] = None
+    is_prerelease: bool = False
+    is_postrelease: bool = False
+    is_devrelease: bool = False
+    is_debug: bool = False
+    version: Optional[Version] = None
+    architecture: Optional[str] = None
+    comes_from: Optional["PathEntry"] = None
+    executable: Optional[Union[str, WindowsPath, Path]] = None
+    company: Optional[str] = None
+    name: Optional[str] = None
+
+    class Config:
+        validate_assignment = True
+        arbitrary_types_allowed = True
+        allow_mutation = True
+        include_private_attributes = True
+        # keep_untouched = (cached_property,)
 
     def __getattribute__(self, key):
-        result = super(PythonVersion, self).__getattribute__(key)
+        result = super().__getattribute__(key)
         if key in ["minor", "patch"] and result is None:
-            executable = None  # type: Optional[str]
+            executable = None
             if self.executable:
                 executable = self.executable
             elif self.comes_from:
@@ -377,7 +348,7 @@ class PythonVersion(object):
                 instance_dict = self.parse_executable(executable)
                 for k in instance_dict.keys():
                     try:
-                        super(PythonVersion, self).__getattribute__(k)
+                        super().__getattribute__(k)
                     except AttributeError:
                         continue
                     else:
@@ -386,8 +357,7 @@ class PythonVersion(object):
         return result
 
     @property
-    def version_sort(self):
-        # type: () -> Tuple[int, int, Optional[int], int, int]
+    def version_sort(self) -> tuple[int, int, int | None, int, int]:
         """
         A tuple for sorting against other instances of the same class.
 
@@ -417,13 +387,11 @@ class PythonVersion(object):
         )
 
     @property
-    def version_tuple(self):
-        # type: () -> Tuple[int, Optional[int], Optional[int], bool, bool, bool]
+    def version_tuple(self) -> tuple[int, int, int, bool, bool, bool]:
         """
         Provides a version tuple for using as a dictionary key.
 
         :return: A tuple describing the python version meetadata contained.
-        :rtype: tuple
         """
 
         return (
@@ -437,21 +405,20 @@ class PythonVersion(object):
 
     def matches(
         self,
-        major=None,  # type: Optional[int]
-        minor=None,  # type: Optional[int]
-        patch=None,  # type: Optional[int]
-        pre=False,  # type: bool
-        dev=False,  # type: bool
-        arch=None,  # type: Optional[str]
-        debug=False,  # type: bool
-        python_name=None,  # type: Optional[str]
-    ):
-        # type: (...) -> bool
+        major: int | None = None,
+        minor: int | None = None,
+        patch: int | None = None,
+        pre: bool = False,
+        dev: bool = False,
+        arch: str | None = None,
+        debug: bool = False,
+        python_name: str | None = None,
+    ) -> bool:
         result = False
         if arch:
             own_arch = self.get_architecture()
             if arch.isdigit():
-                arch = "{0}bit".format(arch)
+                arch = f"{arch}bit"
         if (
             (major is None or self.major == major)
             and (minor is None or self.minor == minor)
@@ -469,20 +436,16 @@ class PythonVersion(object):
             result = True
         return result
 
-    def as_major(self):
-        # type: () -> PythonVersion
-        self_dict = attr.asdict(self, recurse=False, filter=_filter_none).copy()
-        self_dict.update({"minor": None, "patch": None})
-        return self.create(**self_dict)
+    def as_major(self) -> PythonVersion:
+        self.minor = None
+        self.patch = None
+        return self
 
-    def as_minor(self):
-        # type: () -> PythonVersion
-        self_dict = attr.asdict(self, recurse=False, filter=_filter_none).copy()
-        self_dict.update({"patch": None})
-        return self.create(**self_dict)
+    def as_minor(self) -> PythonVersion:
+        self.patch = None
+        return self
 
-    def as_dict(self):
-        # type: () -> Dict[str, Union[int, bool, Version, None]]
+    def as_dict(self) -> dict[str, int | bool | Version | None]:
         return {
             "major": self.major,
             "minor": self.minor,
@@ -495,8 +458,7 @@ class PythonVersion(object):
             "company": self.company,
         }
 
-    def update_metadata(self, metadata):
-        # type: (Dict[str, Union[str, int, Version]]) -> None
+    def update_metadata(self, metadata) -> None:
         """
         Update the metadata on the current :class:`pythonfinder.models.python.PythonVersion`
 
@@ -514,9 +476,7 @@ class PythonVersion(object):
                 setattr(self, key, metadata[key])
 
     @classmethod
-    @lru_cache(maxsize=1024)
-    def parse(cls, version):
-        # type: (str) -> Dict[str, Union[str, int, Version]]
+    def parse(cls, version) -> dict[str, str | int | Version]:
         """
         Parse a valid version string into a dictionary
 
@@ -527,7 +487,6 @@ class PythonVersion(object):
 
         :param str version: A valid version string
         :return: A dictionary with metadata about the specified python version.
-        :rtype: dict
         """
 
         if version is None:
@@ -537,8 +496,7 @@ class PythonVersion(object):
             raise ValueError("Not a valid python version: %r" % version)
         return version_dict
 
-    def get_architecture(self):
-        # type: () -> str
+    def get_architecture(self) -> str:
         if self.architecture:
             return self.architecture
         arch = None
@@ -552,8 +510,9 @@ class PythonVersion(object):
         return self.architecture
 
     @classmethod
-    def from_path(cls, path, name=None, ignore_unsupported=True, company=None):
-        # type: (Union[str, PathEntry], Optional[str], bool, Optional[str]) -> PythonVersion
+    def from_path(
+        cls, path, name=None, ignore_unsupported=True, company=None
+    ) -> PythonVersion:
         """
         Parses a python version from a system path.
 
@@ -566,13 +525,7 @@ class PythonVersion(object):
         :param bool ignore_unsupported: Whether to ignore or error on unsupported paths.
         :param Optional[str] company: The company or vendor packaging the distribution.
         :return: An instance of a PythonVersion.
-        :rtype: :class:`~pythonfinder.models.python.PythonVersion`
         """
-
-        from .path import PathEntry
-
-        if not isinstance(path, PathEntry):
-            path = PathEntry.create(path, is_root=False, only_python=True, name=name)
         from ..environment import IGNORE_UNSUPPORTED
 
         ignore_unsupported = ignore_unsupported or IGNORE_UNSUPPORTED
@@ -602,14 +555,12 @@ class PythonVersion(object):
         instance_dict.update(
             {"comes_from": path, "name": name, "executable": path.path.as_posix()}
         )
-        return cls(**instance_dict)  # type: ignore
+        return cls(**instance_dict)
 
     @classmethod
-    @lru_cache(maxsize=1024)
-    def parse_executable(cls, path):
-        # type: (str) -> Dict[str, Optional[Union[str, int, Version]]]
-        result_dict = {}  # type: Dict[str, Optional[Union[str, int, Version]]]
-        result_version = None  # type: Optional[str]
+    def parse_executable(cls, path) -> dict[str, str | int | Version | None]:
+        result_dict = {}
+        result_version = None
         if path is None:
             raise TypeError("Must pass a valid path to parse.")
         if not isinstance(path, str):
@@ -626,19 +577,16 @@ class PythonVersion(object):
         return result_dict
 
     @classmethod
-    def from_windows_launcher(cls, launcher_entry, name=None, company=None):
-        # type: (Environment, Optional[str], Optional[str]) -> PythonVersion
+    def from_windows_launcher(
+        cls, launcher_entry, name=None, company=None
+    ) -> PythonVersion:
         """Create a new PythonVersion instance from a Windows Launcher Entry
 
         :param launcher_entry: A python launcher environment object.
         :param Optional[str] name: The name of the distribution.
         :param Optional[str] company: The name of the distributing company.
         :return: An instance of a PythonVersion.
-        :rtype: :class:`~pythonfinder.models.python.PythonVersion`
         """
-
-        from .path import PathEntry
-
         creation_dict = cls.parse(launcher_entry.info.version)
         base_path = ensure_path(launcher_entry.info.install_path.__getattr__(""))
         default_path = base_path / "python.exe"
@@ -665,39 +613,40 @@ class PythonVersion(object):
         return py_version
 
     @classmethod
-    def create(cls, **kwargs):
-        # type: (...) -> PythonVersion
+    def create(cls, **kwargs) -> PythonVersion:
         if "architecture" in kwargs:
             if kwargs["architecture"].isdigit():
-                kwargs["architecture"] = "{0}bit".format(kwargs["architecture"])
+                kwargs["architecture"] = "{}bit".format(kwargs["architecture"])
         return cls(**kwargs)
 
 
-@attr.s
-class VersionMap(object):
-    versions = attr.ib(
-        factory=defaultdict
-    )  # type: DefaultDict[Tuple[int, Optional[int], Optional[int], bool, bool, bool], List[PathEntry]]
+class VersionMap(FinderBaseModel):
+    versions: DefaultDict[
+        Tuple[int, Optional[int], Optional[int], bool, bool, bool], List[PathEntry]
+    ] = defaultdict(list)
 
-    def add_entry(self, entry):
-        # type: (...) -> None
-        version = entry.as_python  # type: PythonVersion
+    class Config:
+        validate_assignment = True
+        arbitrary_types_allowed = True
+        allow_mutation = True
+        include_private_attributes = True
+        # keep_untouched = (cached_property,)
+
+    def add_entry(self, entry) -> None:
+        version = entry.as_python
         if version:
             _ = self.versions[version.version_tuple]
             paths = {p.path for p in self.versions.get(version.version_tuple, [])}
             if entry.path not in paths:
                 self.versions[version.version_tuple].append(entry)
 
-    def merge(self, target):
-        # type: (VersionMap) -> None
+    def merge(self, target) -> None:
         for version, entries in target.versions.items():
             if version not in self.versions:
                 self.versions[version] = entries
             else:
                 current_entries = {
-                    p.path
-                    for p in self.versions[version]  # type: ignore
-                    if version in self.versions
+                    p.path for p in self.versions[version] if version in self.versions
                 }
                 new_entries = {p.path for p in entries}
                 new_entries -= current_entries
